/*
	$Workfile:$
	$Header:$

	Copyright  1997-2000 MDS SCIEX.
	All rights reserved.

	@doc
	@module srm.cpp - Contains the main functionality |
	
	This program monitors registry activity and generates a 
	version 4 .REG file.  Arguments to the program are the executable
	file you wish to monitor while self-registering.  .EXE, .OCX and .DLL's
	are supported.

	The program uses work (REGMON) from Mark Russinovich and Bryce Cogswell 
	at www.sysinternals.com to hook the registry api using a custom 
	device driver.  It only supports NT and 2000 for now.  Currently version 4.32 of Regmon is supported
	
	Self Registration Monitor (SRM) was written by Friedrich Brunzema - questions/comments bug fixes
	suggestions for improvement should go to brunzefb@yahoo.com


	Revision History:
	2001 Feb 19  Added support for version 4.3 of Regsys.sys

	$Log$ 
*/


#include "stdafx.h"
#include "getopt.h"
#include "srm.h"
#undef NO_SHLWAPI_STRFCNS
#include "shlwapi.h"
#include <winioctl.h>
extern "C"
{
	#include "ioctlcmd.h"
	#include "regmon.h"
}




//globals
TCHAR								g_szFile[_MAX_PATH];	// Path to ComServer
TCHAR								g_szOutFile[_MAX_PATH];	// Path RegFile
bool								g_fIgnoreCom;			// Ignore COM .reg stuff if True
bool								g_fDll;					// true if its a DLL
bool								g_fDDLoaded;			// true if Device driver loaded

HANDLE								g_SysHandle	= INVALID_HANDLE_VALUE;		// System handle
FILTER								g_FilterDefinition;		// defined filter
char								g_Stats[ LOGBUFSIZE ];	// Buffer into which driver can copy statistics
DWORD								g_StatsLen;				// Current fraction of buffer filled
TCHAR								g_szImgName[_MAX_PATH]; // name of the image
bool								g_fWritten;				// did we write to .reg file
bool								g_fRegEdit;				// is it regedit compatible?

/*
	Name:				main
	@func
	Description:
		main entry point into the program
	@parm 				int	 | argv	 | arg count
	@parm				TCHAR* | argv[] | arguments as strings
	@parm				TCHAR* | envp[] | environment passed in
	@rdesc				void - n/a
*/
int main(int argc, char* argv[])
{
	BOOL		fOK;			// are we OK
	TCHAR*		p;				// temp ptr
			
	// init
	g_fDll = false;
	g_fWritten = false;
	g_fRegEdit = false;
	g_fIgnoreCom = false;
	ZeroMemory(g_szFile, sizeof(TCHAR) * _MAX_PATH);
	ZeroMemory(g_szOutFile, sizeof(TCHAR) * _MAX_PATH);
	ZeroMemory(g_szImgName, sizeof(TCHAR) * _MAX_PATH);

	// show a header	
	_ftprintf(stdout,_T("\nSelf Registation Monitor (SRM) Version 1.3\n"));
	_ftprintf(stdout,_T("Copyright (c)Friedrich Brunzema 2000-2001. All rights reserved.\n"));
	

	// determine version
	// get NT version, abort if not running on NT
	if( GetVersion() >= 0x80000000 )
	{
		_ftprintf(stdout,_T("Yikes, this only runs on NT!\n\n"));
		return 1;
	}

	// figure out the command line and setup global variables to reflect user
	// choices
	process_arguments(argc, argv);

	// We are using COM, must inititialize the OLE libraries
	CoInitialize(NULL);

	// try loading the device Driver
	if (!LoadTheDeviceDriver())
	{
		return 1;
	}
	
	// setup the filter
	fOK = SetupFilter();
	
	// Launch the registration process, as everything should be setup to
	// log to the device driver.  Note that if we launch and .EXE and it
	// does not stop, this program will "hang" waiting for the other guy
	// to finish up.
	if (fOK)
	{
		// launch either DLLRegister server, or run with /RegServer flag
		fOK = Run();
	}

	// now we talk to the device driver to get the results
	// and perform filtering to create the .reg file.
	if (fOK)
	{
		AnalyseResults();
	}
	
	// check to see if we did anything
	if (g_fIgnoreCom)
	{
		if (!g_fWritten)
		{
			_ftprintf(stdout,_T("Nothing written to .REG file for %s!\n"), g_szFile);
		}
	}	
	
	// if we wrote, let the user know
	if (g_fWritten)
	{
		p = PathFindFileName(g_szFile);
		PathRemoveExtension(p);
		_tcscat(p, _T(".reg"));
		if (_tcslen(g_szOutFile))
			_ftprintf(stdout,_T("Writing %s.\n"), g_szOutFile );
		else
			_ftprintf(stdout,_T("Writing %s.\n"), p);
	}	
			
	// cleanup
	if (g_fDDLoaded)
	{
		if (!UnloadDeviceDriver( SYS_NAME ) )
		{
			_ftprintf(stdout,_T("Error unloading device driver!\n"));
		}
	}
	


	return 0;
}


/*
	Name:				SetupFilter
	@func
	Description:
		Sets up a filter for the DD with either our process name (srm.exe)
		or the process name of the local server, checking only for RegSetValue()
		calls

	@rdesc	bool,  True if loaded OK, false if not
*/
bool SetupFilter()
{
	TCHAR	szTemp[_MAX_PATH];			// Temp string
	DWORD	nb;							// dummy
	
	// init
	ZeroMemory (&g_FilterDefinition, sizeof (FILTER));
	ZeroMemory (szTemp, _MAX_PATH);

	// set settings to log successful writes
	g_FilterDefinition.logaux  = false;
	g_FilterDefinition.logerror = false;
	g_FilterDefinition.logreads = false;
	g_FilterDefinition.logsuccess = true;
	g_FilterDefinition.logwrites = true;

	// we omit filtering at the process level here, we will do this later
	// from the output
	_tcscpy(g_FilterDefinition.includefilter, _T("*"));
	
	
	// tell the device driver of the new filter settings
	if ( ! DeviceIoControl(	g_SysHandle, IOCTL_REGMON_SETFILTER,
									(PVOID) &g_FilterDefinition, sizeof(g_FilterDefinition), 
									NULL, 0, &nb, NULL ) )
	{
		_ftprintf(stdout,_T("Unable to set filter settings to device driver.\n"));
		return false;
	}
	
	// Have driver zero information
	if ( ! DeviceIoControl(	g_SysHandle, IOCTL_REGMON_ZEROSTATS,
									NULL, 0, NULL, 0, &nb, NULL ) )
	{
		_ftprintf(stdout,_T("Unable to clear data inside of driver.\n"));
		return false;
	}

	// tell device driver to start working
	if ( ! DeviceIoControl(	g_SysHandle, IOCTL_REGMON_HOOK,
											NULL, 0, NULL, 0, &nb, NULL ) ) 
	{
		_ftprintf(stdout,_T("Unable to start capture of data.\n"));
		return false;
	}

	return true;
}

/*
	Name:				Run
	@func
	Description:
		Runs the local server or calls DLLRegister server in the DLL or OCX
	@rdesc	bool,  True if loaded OK, false if not
*/
bool Run()
{
	HANDLE				h;					// library handle
	HRESULT				hr;					// HRESULT of the DLLReg call
	PROCESS_INFORMATION pi;					// process information
	STARTUPINFO			si;					// startup info
	BOOL				bRet;				// return from CreateProcess

	if (g_fDll)		
	{
		// its a DLL or OCX, so we must call DLL register server
		typedef HRESULT (__stdcall * LPFNDLLFUNC1)(void);
		LPFNDLLFUNC1 RegSvr;

		// to start out, we load the library
		h = LoadLibrary(g_szFile);
		if (NULL == h)
		{
			_ftprintf(stdout,_T("LoadLibrary Failed.\n"));
			return false;
		}	
		
		// get the address of the procedure
		RegSvr = (LPFNDLLFUNC1)GetProcAddress((HINSTANCE)h, "DllRegisterServer");
		if (NULL == RegSvr)
		{
			_ftprintf(stdout,_T("GetProcAddress of DllRegisterServer failed.\n"));
			FreeLibrary((HMODULE) h);
			return false;
		}	
		
		// make the call
		hr = RegSvr();	
		if (!SUCCEEDED(hr))
		{
			_ftprintf(stdout,_T("Call to DllRegisterServer failed on HRESULT.\n"));
			FreeLibrary((HMODULE) h);
			return false;
		}
		
		// cleanup
		FreeLibrary((HMODULE) h);
		
	}
	else	// we've got an .EXE that we need to run
	{
		// get ready
		ZeroMemory(&si, sizeof(si));
		si.cb = sizeof(si);
		si.wShowWindow = SW_NORMAL;
		si.dwFlags = STARTF_USESHOWWINDOW;

		// launch him
		bRet = CreateProcess (g_szFile,
						_T(" /REGSERVER"),
						NULL,
						NULL,
						FALSE,
						NORMAL_PRIORITY_CLASS,
						NULL,
						NULL,	
						&si,
						&pi);

		// did we launch OK?
		if (bRet)
		{
			// wait until we are done with  process
			WaitForSingleObject(pi.hProcess, INFINITE);
		}
		else
		{
			_ftprintf(stdout,_T("Error launching Executable.\n"));
			return false;
		}
		
		// cleanup
		CloseHandle(pi.hProcess);
	}
	
	// everything peachy
	return true;
}	


/*
	Name:				LoadTheDeviceDriver()
	@func
	Description:
		Loads the device driver
	@rdesc	bool,  True if loaded OK, false if not
*/
bool LoadTheDeviceDriver()
{
	DWORD	dwError;					// error id
	TCHAR	systemRoot[_MAX_PATH];		// pointer to the system root
	TCHAR   driverPath[_MAX_PATH];		// path to the driver
	DWORD	dwVersionNumber;			// version number
	DWORD	nb;							// number

	// get the system directory
	if( !GetEnvironmentVariable( "SYSTEMROOT", systemRoot, sizeof(systemRoot))) 
	{
		return false;
	}
	
	// create a full path to the driver
	_stprintf( driverPath, _T("%s\\system32\\drivers\\%s"), systemRoot, SYS_FILE );

	// load it <winerror.h>
	if ( ! LoadDeviceDriver( SYS_NAME, driverPath, &g_SysHandle, &dwError ) )
	{
		_ftprintf(stdout,_T("Error Opening %s.SYS\n"), SYS_NAME);
		_ftprintf(stdout,_T("Regmon.sys needs to be installed in the \\winnt\\system32\\drivers folder\n"));
		_ftprintf(stdout,_T("Regmon.sys available from http://www.sysinternals.com/ntw2k/source/regmon.shtml"));
		_ftprintf(stdout,_T("Also make sure that REGMON.EXE is not running\n"));
		return false;
	}

	// set the flag to unload at the end
	g_fDDLoaded = true;
	

	// Correct driver version?
	
	if ( ! DeviceIoControl(	g_SysHandle, IOCTL_REGMON_VERSION,
							NULL, 0, &dwVersionNumber, sizeof(DWORD), &nb, NULL ) ||
			dwVersionNumber != REGMON_VERSION )
	{
		_ftprintf(stdout,_T("Wrong Device Driver Version of REGSYS.sys!\n"));
	}
	
	return true;
}


/*
	Name:				process_arguments
	@func
	Description:
		This function analyzes the command line arguments
	@parm 				int	 | argc	 | argument count
	@parm				TCHAR* | argv[] | command line argument
	@rdesc				n/a 
*/
void process_arguments(int argc, TCHAR *argv[])
{
    TCHAR       c;                          /* flag specifed, used w/getopt  */
	int			optind;						// option index
	TCHAR*		pOptArg;					// optional arg
	

	// init
	optind = 0;

    /*
     * GETOPT is an easy way to parse command line arguments
     * Each parameter which can have a flag is passed in the third argument
     * to getopt.  Getopt returns the character of the flag on the command
     * line and sets optarg to point to the value associated with the flag.
     * optind is the index of the argument that getopt is currently processing.
     */
    while (optind != argc) 
	{
        c = GetOption(argc, argv, _T("?iro:"), &pOptArg);
        switch (c) 
		{
			case 0:			// reached the end of the list
				// did we at least get a filename?  if not, kill ourselves
				if (_tcslen(g_szFile) == 0)
				{
					show_info();
					exit(1);					
				}
				break;

			case 1:			// pOptArg has our non optional argument
				// do some sanity checks here
				VerifyPath(pOptArg);
				break;

			case _T('?'):	// user needs help
				show_info();
				exit(1);
				break;
        
			case _T('i'):	// sample
				g_fIgnoreCom = true;
				break;
	
			case _T('o'):	// tic file name
				_tcscpy(g_szOutFile, pOptArg);
				break;
			
			case _T('r'):	// regedit compatible output
				g_fRegEdit = true;
				break;

			default:
				_ftprintf(stdout, _T("Invalid flag.  Use srm -? for usage\n"));
				exit(1);
        }
		++optind;
    }
}


/*
	Name:				show_info	
	@func
	Description:
		This function shows usage information
	
	@rdesc				n/a
*/
void show_info()
{
	// show dump utility usage information
	_ftprintf(stdout,	
		_T("\nMonitors Self-Registration Activity and creates a .REG file\n\n")
		_T("usage: srm [/i] [/o output .Reg filename] ComServerName \n")
		_T("       srm [-?] \n\n")
		_T("  ComServerName Specifies the ComServer inproc or localserver to monitor.\n\n")
		_T("  /i            Ignore any COM related information\n")
		_T("  /o            Specifies the new filename for the output .REG file.\n")
		_T("                The default name is the ComServerName with a .REG extension.\n")
		_T("  /r            Regedit compatible output\n")
		_T("  /?            Displays this help message.\n")
			);
}


/*
	Name:				VerifyPath
	@func
	Description:
		Checks for the path of the wiff file to see if its OK
	@parm				TCHAR* | szFile | Path to wiff file
	@rdesc				n/a 
*/
void VerifyPath(LPTSTR szFile)
{
	TCHAR		szTemp[_MAX_PATH];				// temp var
	TCHAR*		p;								// a pointer

	// init
	ZeroMemory(szTemp, _MAX_PATH);

	// check to see if the file exists
	if (!PathFileExists(szFile))
	{
		_ftprintf(stdout, _T("The file %s does not exist!\n"), szFile);
		exit(1);
	}

	// if so, copy it to our file buffer
	_tcscpy(g_szFile, szFile);

	// make it lower case
	_tcslwr(g_szFile);

	// check what type we've got
	if (_tcscmp(PathFindExtension(g_szFile), _T(".ocx"))==0 ||
		_tcscmp(PathFindExtension(g_szFile), _T(".dll"))==0	)
	{
		g_fDll = true;
	}
	else if (_tcscmp(PathFindExtension(g_szFile), _T(".exe"))==0)
	{
		g_fDll = false;
	}
	else
	{
		_ftprintf(stdout, _T("This utility can only monitor self-registration of .dll, .ocx and .exe files\n"));
		exit(1);
	}

	// if we're a DLL, our parent process name is srm.exe
	if (g_fDll)
	{
		_tcscpy(g_szImgName, _T("srm.exe"));
	}
	else
	{
		// extract the image name without the path, make it lower
		_tcscpy(szTemp, g_szFile);
		p = PathFindFileName(szTemp);
		_tcslwr(p);
		_tcscpy(g_szImgName, p);
	}
}



/*
	Name:				AnalyseResults
	@func
	Description:
		Gets stats from driver, analyzes and writes .reg file
			
	@rdesc				n/a 
*/
void AnalyseResults()
{
    PNTENTRY	ntptr;					// pointer to the entry
	PCHAR		Buffer;					// pointer to the buffer
	DWORD		BufLen;					// length of buffer
	ULONG		len;					// length
	BOOLEAN		isfirst = TRUE;			// is it first?
	FILE*		fh;						// file pointer
	TCHAR		szTemp[_MAX_PATH];		// temp string
	TCHAR*		p;						// pointer

	// init
	g_StatsLen = 0;
	ZeroMemory(szTemp, _MAX_PATH);

	
	// Have driver fill Stats buffer with information
	if ( ! DeviceIoControl(	g_SysHandle, IOCTL_REGMON_GETSTATS,
							NULL, 0, &g_Stats, sizeof (g_Stats),
							&g_StatsLen, NULL ) )
	{
		_ftprintf(stdout,_T("Could not get statistics from the driver\n"));
		return;
	}
	
	// sanity check
	if ( g_StatsLen == 0 )
	{
		return;
	}

	// get a filename
	_tcscpy(szTemp, g_szFile);
	p = PathFindFileName(szTemp);
	PathRemoveExtension(p);
	_tcscat(p, _T(".reg"));

	// open it
	if (_tcslen(g_szOutFile) == 0)
	{
		fh = _tfopen(p, _T("wt"));
	}
	else
	{
		fh = _tfopen(g_szOutFile, _T("wt"));
	}
	if (NULL == fh)
	{
		_ftprintf(stdout, _T("Could not open %s for write\n"), p);
		return;
	}

	// start a header
	_ftprintf(fh, _T("REGEDIT4\n\n"));

	// avoid making too many changes to the code
	Buffer = (PCHAR) g_Stats;
	BufLen = g_StatsLen;

	// loop over the entries in the buffer
	for ( ntptr = (PNTENTRY)Buffer; (char *)ntptr < min(Buffer+BufLen,Buffer + LOGBUFSIZE); )
	{
		len = _tcslen(ntptr->text);
		len += 4; 
		len &= 0xFFFFFFFC; // +1 for null-terminator +3 for 32bit alignment

		// truncate if necessary
		if( len > MAXITEMLENGTH - 1 ) ntptr->text[MAXITEMLENGTH-1] = 0;
		isfirst = FALSE;

		// parse the line (enable next line for for debugging SRM
		//_ftprintf(stdout, _T("%s\n"), ntptr->text);
	
		ParseLine(fh, ntptr->text);
				
		ntptr = (PNTENTRY)(ntptr->text + len);
	}

	// and cleanup
	fclose(fh);
}



/*
	Name:				ParseLine
	@func
	Description:
		Parses stats from the driver and writes .reg file
	@parm	FILE* | fh | File handle to the .REG file.			
	@parm	TCHAR* | szLine | Line of text
	@rdesc				n/a 
*/
void ParseLine(FILE* fh, TCHAR* szLine)
{
	TCHAR			szIn[1024];			// temp var
	TCHAR*			p;					// pointer to substring
	TCHAR			szTemp[512];		// temp var
	TCHAR			szOut[512];			// temp var
	bool			fSetValue;			// action = SetValue
	bool			fCreateKey;			// action = CreateKey
	TCHAR			szValue[_MAX_PATH];	// the value name of the key
	TCHAR*			pT;					// temp pointer
	DWORD			dw;					// temp DWORD
	BOOL			fIsHKCR;			// is it HKCR?
		
	// init
	ZeroMemory(szIn, 1024);
	ZeroMemory(szTemp, 512);
	fIsHKCR = false;

	// make a copy
	_tcscpy(szIn, szLine);

	// grab the process name
	p =	_tcstok(szIn, _T(":\t"));	

	// is it for us?, bug: regmon truncates to 16 chars 
	// thanks to Graeme Foster for finding this!
	if (_tcsnicmp(p, g_szImgName, 15) != 0)
		return;

	// next token = PID
	p = _tcstok(NULL, _T(":\t"));

	// next token = Action
	p = _tcstok(NULL, _T(":\t"));

	// decode action
	fSetValue = !_tcscmp(p, _T("SetValue"));
	fCreateKey = !_tcscmp(p, _T("CreateKey")); 
	
	// only interested in  SetValue and CreateKey
	if (!fSetValue && !fCreateKey)
		return;
	
	// next token = Registry key, encoded  HKLM\sdfs
	p = _tcstok(NULL, _T(":\t"));
	

	// expand the encoded reg key bases
	if (!_tcsncmp(p, _T("HKLM"), _tcslen(_T("HKLM"))))
	{
		_stprintf(szTemp, _T("[HKEY_LOCAL_MACHINE%s]"), &p[_tcslen("HKLM")]);
	}
	else if(!_tcsncmp(p, _T("HKCU"), _tcslen(_T("HKCU"))))
	{
		_stprintf(szTemp, _T("[HKEY_CURRENT_USER%s]"), &p[_tcslen("HKCU")]);
	}
	else if(!_tcsncmp(p, _T("HKCC"), _tcslen(_T("HKCC"))))
	{
		_stprintf(szTemp, _T("[HKEY_CURRENT_CONFIG%s]"), &p[_tcslen("HKCC")]);
	}
	else if(!_tcsncmp(p, _T("HKCR"), _tcslen(_T("HKCR"))))
	{
		_stprintf(szTemp, _T("[HKEY_CLASSES_ROOT%s]"), &p[_tcslen("HKCR")]);
		fIsHKCR = true;
	}
	else if(!_tcsncmp(p, _T("HKU"), _tcslen(_T("HKU"))))
	{
		_stprintf(szTemp, _T("[HKEY_LOCAL_MACHINE%s]"), &p[_tcslen("HKU")]);
	}
	else
	{
		_ftprintf(stdout, _T("Error during parse, line=%s"), szLine);
		return;
	}

	// is it just a create key, deal with it	
	if (fCreateKey)
	{
		if ( !(fIsHKCR && g_fIgnoreCom) )
		{
			_ftprintf(fh, _T("\n%s\n"), szTemp);
			_ftprintf(fh, _T("@=\"\"\n\n"));
			g_fWritten = true;
		}
		return;
	}

	// next token = SUCCESS/FAIL
	p = _tcstok(NULL, _T(":\t"));
	
	// next token = value
	p = _tcstok(NULL, _T("\t"));
	
	// is it a SetValue, because we tested earlier for either CreateKey or SetValue
	// figure out the key name an value name
	pT = StrRChr(szTemp, NULL, _T('\\'));
	if (NULL == pT)
		return;
	
	// if the trailer is "Default, use the @ otherwise it becomes the Key value name
	if (!_tcscmp(pT + 1, _T("(Default)]")) )
	{
		_tcscpy( szValue, _T("@"));
	}
	else
	{
		// get the key name
		_tcscpy( szValue, pT + 1);
		// kill the terminating ']'
		szValue[_tcslen(szValue) - 1] = 0;
		if (!g_fRegEdit)
		{
			// add quotes
			_tcscpy(szOut, szValue);
			_stprintf(szValue, _T("\"%s\""), szOut);	
		}
		
	}

	// close the right bracket of the main key
	*pT = _T(']');
	*(pT + 1) = 0;

	// this code is some special code that we use to monitor a user specific key value in HKCR
	// if its present, we want to log it, even though we specify not to use other HKCR values
	int fIsMode;
	fIsMode = _tcsicmp(szValue, _T("\"ModeFlagForWSSave\""));
	// even if we are not showing COM registry info, we got to take this one
	if (fIsHKCR && g_fIgnoreCom)
	{
		if (fIsMode!=0)
			return;
	}
	// end special code

	// print the key name
	_ftprintf(fh, _T("\n%s\n"), szTemp);
	g_fWritten = true;
	
	if (p[0] == _T('\"'))	// string
	{
		_ftprintf(fh, _T("%s=%s\n"), szValue, p);
	}
	else if (p[1] == _T('x')) // dword value
	{
		// regedit wants a special format, so do it
		_stscanf(p, _T("%x"), &dw);
		_ftprintf(fh, _T("%s=dword:%08x\n"), szValue, dw);
	}
	else if (_tcslen(p) >=3 && p[2] == _T(' '))  // hex binary values
	{
		// limitation of the device driver
		_ftprintf(stdout, _T("Sorry, binary reg values not supported.\n"));
		_ftprintf(fh, _T("@=\"Binary Value Missing\"\n"), szValue, p);
	}	
	else
	{
		_ftprintf(stdout, _T("Something went wrong parsing %s.\n"), p);
	}
}
